@@ -7,7 +7,7 @@ import subprocess
 import sys
 import threading
 import traceback
-from asyncio.exceptions import IncompleteReadError
+from asyncio.exceptions import CancelledError, IncompleteReadError, TimeoutError
 from pathlib import Path
 from typing import List, Optional
 
@@ -18,6 +18,14 @@ from websockets.exceptions import ConnectionClosedError
 from pywry import pywry
 
 __all__ = ["PyWry", "BackendFailedToStart"]
+Websocket_Error = (
+    TimeoutError,
+    OSError,
+    CancelledError,
+    socket.gaierror,
+    ConnectionClosedError,
+    IncompleteReadError,
+)
 
 
 class BackendFailedToStart(Exception):
@@ -65,16 +73,9 @@ class PyWry:
         self._is_closed: asyncio.Event = asyncio.Event()
         self._is_closed.set()
 
-        port = self.get_clean_port()
-
-        for host in ["host.docker.internal", socket.gethostname()]:
-            try:
-                self.host = socket.gethostbyname(host)
-                break
-            except (socket.gaierror, socket.herror, Exception):
-                self.host = "localhost"
-
-        self.url = f"ws://{self.host}:{port}"
+        self.port = self.get_clean_port()
+        self.host = "localhost"
+        self.url = f"ws://{self.host}:{self.port}"
 
         atexit.register(self.close)
 
@@ -84,6 +85,23 @@ class PyWry:
         else:
             self.procs.clear()
 
+    async def get_valid_host(self, port: int) -> str:
+        """Get a valid host that connects to the backend."""
+        try_hosts = [
+            "127.0.1.0",
+            "0.0.0.0",
+            "host.docker.internal",
+            "localhost",
+            socket.gethostbyname(socket.gethostname()),
+        ]
+
+        for host in try_hosts:
+            try:
+                if await self.send_test(host, port):
+                    return host
+            except Websocket_Error:
+                pass
+
     def send_html(self, html_str: str = "", html_path: str = "", title: str = ""):
         """Send html to backend.
 
@@ -125,10 +143,23 @@ class PyWry:
             self.print_debug()
             await self.handle_start()
 
-    async def send_test(self):
+    async def send_test(self, host: str, port: int):
         """Send data to the backend."""
-        async with connect(self.url) as websocket:
-            await websocket.send("<test>")
+        async with connect(
+            f"ws://{host}:{port}",
+            open_timeout=6,
+            timeout=1,
+            ssl=None,
+        ) as websocket:
+            while True:
+                try:
+                    await websocket.send("<test>")
+                    response = await websocket.recv()
+                    if response == "SUCCESS":
+                        return True
+                    return False
+                except Websocket_Error:
+                    return False
 
     def get_clean_port(self) -> str:
         """Get a clean port to use for the backend."""
@@ -152,6 +183,7 @@ class PyWry:
                     self.runner = None
                     self._is_started.clear()
                     self._is_closed.set()
+                    self.port = port
                     self.url = f"ws://{self.host}:{port}"
 
             kwargs = {}
@@ -204,6 +236,9 @@ class PyWry:
 
         await asyncio.sleep(1)
 
+        self.host = await self.get_valid_host(self.port)
+        self.url = f"ws://{self.host}:{self.port}"
+
         try:
             async with connect(
                 self.url,
@@ -278,4 +313,5 @@ class PyWry:
         if not reset:
             for process in [p for p in self.procs if p.is_running()]:
                 for child in process.children(recursive=True):
-                    child.kill()
+                    if child.is_running():
+                        child.kill()
